// Copyright (C) 2021 Bosutech XXI S.L.
//
// nucliadb is offered under the AGPL v3.0 and as commercial software.
// For commercial licensing, contact us at info@nuclia.com.
//
// AGPL:
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.
//

pub mod grpc;
pub mod shards;

use tokio_util::sync::CancellationToken;
use tracing::debug;

use crate::{Settings, grpc_server::GrpcServer};

pub async fn run(settings: Settings, shutdown: CancellationToken) -> anyhow::Result<()> {
    let service = grpc::ApiServer::new(&settings).into_router();
    let server = GrpcServer::new("0.0.0.0:10000").await?;
    debug!("Running API at port {}", server.port()?);
    server.serve(service, shutdown).await?;

    Ok(())
}
